import errno
import json
import os
from argparse import Namespace
from configparser import ConfigParser
from typing import List, Set, Any, Dict, Callable, Optional

from wordfence.logging import log
from .config_items import Context, ConfigItemDefinition, \
    CanonicalValueExtractorInterface, not_set_token, \
    get_config_map_for_subcommand, subcommand_module_map, ReferenceToken
from .defaults import INI_DEFAULT_PATH


valid_contexts: Set[Context] = {Context.ALL, Context.CONFIG}


class IniCanonicalValueExtractor(CanonicalValueExtractorInterface):

    def __init__(self, config_section_name):
        self.config_section_name = config_section_name

    def is_valid_source(self, source: Any) -> bool:
        return isinstance(source, ConfigParser)

    def get_canonical_value(self, definition: ConfigItemDefinition,
                            source: ConfigParser) -> Any:
        self.assert_is_valid_source(source)

        if definition.has_separator():
            # always return separated values as a string
            value = source.get(self.config_section_name,
                               definition.property_name,
                               fallback=not_set_token)
        elif definition.get_value_type() == bool:
            value = source.getboolean(self.config_section_name,
                                      definition.property_name,
                                      fallback=not_set_token)
        elif definition.get_value_type() == int:
            value = source.getint(self.config_section_name,
                                  definition.property_name,
                                  fallback=not_set_token)
        elif isinstance(definition.get_value_type(), Callable):
            value = source.get(self.config_section_name,
                               definition.property_name,
                               fallback=not_set_token)
            # convert using the type method
            value = value if isinstance(value, ReferenceToken) else (
                (definition.get_value_type())(value))
        elif definition.get_value_type() != str:
            raise ValueError(
                "Only string, bool, and int types are currently known to the "
                "INI parser")
        else:
            value = source.get(self.config_section_name,
                               definition.property_name,
                               fallback=not_set_token)
        if isinstance(value, str) and definition.has_separator():
            value = value.split(definition.meta.separator)
            if definition.get_value_type() == int:
                value = [int(string_int) for string_int in value]
            elif definition.get_value_type() != str:
                raise ValueError(
                    "INI files currently support lists of strings and ints, no"
                    " other types")
        return value


def get_definitions(cli_values: Namespace) -> Dict[str, ConfigItemDefinition]:
    return get_config_map_for_subcommand(cli_values.subcommand)


def get_config_section_name(cli_values: Namespace) -> str:
    return subcommand_module_map[cli_values.subcommand].CONFIG_SECTION_NAME


def get_ini_value_extractor(
        cli_values: Namespace) -> IniCanonicalValueExtractor:
    return IniCanonicalValueExtractor(get_config_section_name(cli_values))


def get_ini_path(cli_values: Namespace) -> str:
    if 'configuration' not in cli_values or not isinstance(
            cli_values.configuration, str):
        path = INI_DEFAULT_PATH
    else:
        path = cli_values.configuration
    return os.path.expanduser(path)


def load_ini(cli_values) -> (ConfigParser, Optional[str]):
    config = ConfigParser()
    ini_path = get_ini_path(cli_values)
    try:
        with open(ini_path) as file:
            config.read_file(file)
    except OSError as e:
        if e.errno == errno.EACCES:
            raise PermissionError(
                f"The current user cannot read the config file: "
                f"{json.dumps(get_ini_path(cli_values))}") from e
        elif e.errno != errno.ENOENT:
            raise
        # config file does not exist: proceed with default values + CLI values
        log.info(
            f"Config file not found or not readable: "
            f"{json.dumps(get_ini_path(cli_values))}. Merging default config "
            f"values.")
        return (config, None)
    config_section_name = get_config_section_name(cli_values)
    definitions = get_definitions(cli_values)
    all_section_names: List[str] = config.sections()
    all_section_names.append("DEFAULT")
    invalid_settings: bool = False
    for section_name in all_section_names:
        if section_name != config_section_name:
            config.remove_section(section_name)
    # remove values that are in the incorrect context or are entirely unknown
    for property_name, _value in config.items(config_section_name):
        # arguments are stored in the lookup by name (kebab-case), but written
        # out in snake_case in the INI
        key = property_name.replace('_', '-')
        # detect unknown definitions and definitions written in kebab-case
        # instead of snake_case
        if key not in definitions or (
                key == property_name and '-' in property_name):
            log.warning(
                f"Ignoring unknown config setting {json.dumps(property_name)}")
            config.remove_option(config_section_name, key)
            invalid_settings = True
        if key in definitions:
            valid_ini_value: bool = definitions[key].context in valid_contexts
            if not valid_ini_value:
                log.warning(
                    f"Ignoring setting that is not valid in the config file "
                    f"context: {json.dumps(definitions[key].name)}.")
                invalid_settings = True
                config.remove_option(config_section_name, key)
                continue
    if invalid_settings:
        log.warning(
            "*** Invalid settings not known to wordfence-cli or that are not "
            "intended for use in INI config files were discarded. ***")

    return (config, ini_path)



from typing import Dict, Any

from .config_definitions import config_definitions

CLI_TITLE = 'Scan files'
CONFIG_SECTION_NAME = 'SCAN'


def get_config_definitions() -> Dict[str, Dict[str, Any]]:
    return config_definitions



import json

from .noc_client import NocClient
from .exceptions import ApiException

from ..intel.signatures import CommonString, Signature, SignatureSet
from ..util.validation import DictionaryValidator, ListValidator, Validator

NOC1_BASE_URL = 'https://noc1.wordfence.com/v2.27/'


class Client(NocClient):

    def get_default_base_url(self) -> str:
        return NOC1_BASE_URL

    def _generate_site_stats(self) -> str:
        return json.dumps({})

    def build_query(self, action: str, base_query: dict = None) -> dict:
        query = super().build_query(action, base_query)
        query['s'] = self._generate_site_stats()
        return query

    def validate_response(self, response, validator: Validator) -> None:
        if isinstance(response, dict) and 'errorMsg' in response:
            raise ApiException(
                    'Error message received in response body',
                    response['errorMsg']
                )
        return super().validate_response(response, validator)

    def get_patterns(self) -> dict:
        patterns = self.request('get_patterns')
        validator = DictionaryValidator({
            'badstrings': ListValidator(str),
            'commonStrings': ListValidator(str),
            'rules': ListValidator(ListValidator({
                0: int,
                1: int,
                2: str,
                3: str,
                4: str,
                5: int,
                6: str,
                7: str,
                8: ListValidator(int)
            })),
            'signatureUpdateTime': int,
            'word1': str,
            'word2': str,
            'word3': str
        })
        self.validate_response(patterns, validator)
        return patterns

    def get_malware_signatures(self) -> SignatureSet:
        patterns = self.get_patterns()
        common_strings = []
        signatures = {}
        for string in patterns['commonStrings']:
            common_strings.append(CommonString(string))
        for record in patterns['rules']:
            if record[5] != 0:
                continue
            signature_id = record[0]
            signatures[signature_id] = Signature(
                signature_id,
                record[2],
                record[7],
                record[3],
                record[8]
            )
            for index in record[8]:
                try:
                    common_strings[index].signature_ids.append(signature_id)
                except IndexError as index_error:
                    raise ApiException(
                            'Response data contains malformed common string '
                            'association'
                        ) from index_error
        return SignatureSet(common_strings, signatures, self.license)

    def ping_api_key(self) -> bool:
        response = self.request('ping_api_key')
        validator = DictionaryValidator({
                'ok': int
            })
        self.validate_response(response, validator)
        return bool(response['ok'])



from .exceptions import ApiException


LICENSE_URL = 'https://www.wordfence.com/products/wordfence-cli/'


class License:

    def __init__(self, key: str):
        self.key = key

    def __eq__(self, other):
        return other.key == self.key


class LicenseRequiredException(ApiException):
    pass


class LicenseSpecific:

    def __init__(self, license: License):
        self.license = license

    def is_compatible_with_license(self, license: License):
        return self.license == license



import logging
from typing import Optional

DEFAULT_LOGGER_NAME = 'wordfence'

logging.basicConfig(format='%(message)s')
log = logging.getLogger(DEFAULT_LOGGER_NAME)
root_log = logging.getLogger()

initial_handler: Optional[logging.Handler] = None


def remove_initial_handler() -> None:
    global initial_handler
    if initial_handler is not None:
        return
    initial_handler = root_log.handlers[0]
    root_log.removeHandler(initial_handler)


def restore_initial_handler(error_if_not_set: bool = False) -> None:
    global initial_handler
    if initial_handler is None:
        if error_if_not_set:
            raise ValueError("Unknown initial handler")
        return
    root_log.addHandler(initial_handler)
    initial_handler = None



import sys
import signal
import os
import logging
from multiprocessing import parent_process
from contextlib import nullcontext
from typing import Any, Optional

from wordfence import scanning, api
from wordfence.api.licensing import LicenseSpecific
from wordfence.scanning import filtering
from wordfence.scanning.scanner import ExceptionContainer
from wordfence.util import caching, updater, pcre
from wordfence.util.io import StreamReader
from wordfence.intel.signatures import SignatureSet
from wordfence.logging import (log, remove_initial_handler,
                               restore_initial_handler)
from wordfence.version import __version__
from .reporting import Report, ReportFormat
from .configure import Configurer
from .progress import ProgressDisplay, ProgressException, reset_terminal


screen_handler: Optional[logging.Handler] = None


def revert_progress_changes() -> None:
    global screen_handler
    if screen_handler:
        log.removeHandler(screen_handler)
    restore_initial_handler()
    reset_terminal()


class ScanCommand:

    CACHEABLE_TYPES = {
            'wordfence.intel.signatures.SignatureSet',
            'wordfence.intel.signatures.CommonString',
            'wordfence.intel.signatures.Signature',
            'wordfence.api.licensing.License'
        }

    def __init__(self, config):
        self.config = config
        self.cache = self._initialize_cache()
        self.license = None
        self.cache.add_filter(self.filter_cache_entry)
        self.cacheable_signatures = None

    def filter_cache_entry(self, value: Any) -> Any:
        if isinstance(value, LicenseSpecific):
            if not value.is_compatible_with_license(self._get_license()):
                raise caching.InvalidCachedValueException(
                        'Incompatible license'
                    )
        return value

    def _get_license(self) -> api.licensing.License:
        if self.license is None:
            if self.config.license is None:
                raise api.licensing.LicenseRequiredException()
            self.license = api.licensing.License(self.config.license)
        return self.license

    def _initialize_cache(self) -> caching.Cache:
        if self.config.cache:
            try:
                return caching.CacheDirectory(
                        os.path.expanduser(self.config.cache_directory),
                        self.CACHEABLE_TYPES
                    )
            except caching.CacheException as e:
                log.warning('Failed to initialize cache directory: ' + str(e))
        return caching.RuntimeCache()

    def filter_signatures(self, signatures: SignatureSet) -> None:
        if self.config.include_signatures:
            for identifier in list(signatures.signatures.keys()):
                if identifier not in self.config.include_signatures:
                    signatures.remove_signature(identifier)
            for identifier in self.config.include_signatures:
                if identifier in signatures.signatures:
                    log.debug(f'Including signature: {identifier}')
                else:
                    log.warning(
                                f'Signature {identifier} was not found and '
                                'could not be included'
                            )
        if self.config.exclude_signatures is not None:
            for identifier in self.config.exclude_signatures:
                if signatures.remove_signature(identifier):
                    log.debug(f'Excluded signature {identifier}')
                else:
                    log.warning(
                            f'Signature {identifier} is not in the existing '
                            'set. It will not be used in the scan.'
                        )
        signature_count = len(signatures.signatures)
        log.debug(f'Filtered signature count: {signature_count}')

    def _get_signatures(self) -> SignatureSet:
        if self.cacheable_signatures is None:
            def fetch_signatures() -> SignatureSet:
                noc1_client = api.noc1.Client(
                        self._get_license(),
                        base_url=self.config.noc1_url
                    )
                return noc1_client.get_malware_signatures()
            self.cacheable_signatures = caching.Cacheable(
                    'signatures',
                    fetch_signatures,
                    86400  # Cache signatures for 24 hours
                )
        signatures = self.cacheable_signatures.get(self.cache)
        self.filter_signatures(signatures)
        return signatures

    def _should_read_stdin(self) -> bool:
        if sys.stdin is None:
            return False
        if self.config.read_stdin is None:
            return not sys.stdin.isatty()
        else:
            return self.config.read_stdin

    def _should_write_stdout(self) -> bool:
        if sys.stdout is None or self.config.output is False:
            return False
        return self.config.output or self.config.output_path is None

    def _get_file_list_separator(self) -> str:
        if isinstance(self.config.file_list_separator, bytes):
            return self.config.file_list_separator.decode('utf-8')
        return self.config.file_list_separator

    def _initialize_file_filter(self) -> filtering.FileFilter:
        filter = filtering.FileFilter()
        has_include_overrides = False
        if self.config.include_files is not None:
            has_include_overrides = True
            for name in self.config.include_files:
                filter.add(filtering.filter_filename(name))
        if self.config.include_files_pattern is not None:
            has_include_overrides = True
            for pattern in self.config.include_files_pattern:
                filter.add(filtering.filter_pattern(pattern))
        if self.config.exclude_files is not None:
            for name in self.config.exclude_files:
                filter.add(filtering.filter_filename(name), False)
        if self.config.exclude_files_pattern is not None:
            for pattern in self.config.exclude_files_pattern:
                filter.add(filtering.filter_pattern(pattern), False)
        if not has_include_overrides:
            filter.add(filtering.filter_php)
            filter.add(filtering.filter_html)
            filter.add(filtering.filter_js)
            if self.config.images:
                filter.add(filtering.filter_images)
        return filter

    def _get_pcre_options(self) -> pcre.PcreOptions:
        return pcre.PcreOptions(
                    caseless=True,
                    match_limit=self.config.pcre_backtrack_limit,
                    match_limit_recursion=self.config.pcre_recursion_limit
                )

    def execute(self) -> int:
        if self.config.purge_cache:
            self.cache.purge()
        if self.config.check_for_update:
            updater.Version.check(self.cache)

        progress = None
        if self.config.progress:
            global screen_handler
            progress = ProgressDisplay(int(self.config.workers))
            screen_handler = progress.get_log_handler()
            if sys.stderr is None or sys.stderr.isatty():
                remove_initial_handler()
            log.addHandler(screen_handler)

        paths = set()
        for argument in self.config.trailing_arguments:
            paths.add(argument)
        options = scanning.scanner.Options(
                paths=paths,
                workers=int(self.config.workers),
                signatures=self._get_signatures(),
                chunk_size=self.config.chunk_size,
                scanned_content_limit=int(self.config.scanned_content_limit),
                file_filter=self._initialize_file_filter(),
                match_all=self.config.match_all,
                pcre_options=self._get_pcre_options(),
                allow_io_errors=self.config.allow_io_errors,
                debug=self.config.debug
            )
        if self._should_read_stdin():
            options.path_source = StreamReader(
                    sys.stdin,
                    self._get_file_list_separator()
                )

        with open(self.config.output_path, 'w') if self.config.output_path \
                is not None else nullcontext() as output_file:
            output_format = ReportFormat(self.config.output_format)
            output_columns = self.config.output_columns
            report = Report(
                    output_format,
                    output_columns,
                    options.signatures,
                    self.config.output_headers
                )
            if self._should_write_stdout():
                if progress:
                    report.add_target(progress.get_output_stream())
                else:
                    report.add_target(sys.stdout)
            if output_file is not None:
                report.add_target(output_file)
            if not report.has_writers():
                log.error(
                        'Please specify an output file using the --output-path'
                        ' option or add --output to write results to standard '
                        'output'
                    )
                return 1
            self.scanner = scanning.scanner.Scanner(options)
            if progress:
                use_log_events = True
            else:
                use_log_events = False
            self.scanner.scan(
                    report.add_result,
                    progress.handle_update if progress is not None else None,
                    progress.scan_finished_handler if progress is not None
                    else None,
                    use_log_events
                )

            if progress:
                progress.end_on_input()
                revert_progress_changes()
                if progress.results_message:
                    print(progress.results_message)
        return 0

    def terminate(self) -> None:
        if hasattr(self, 'scanner') and self.scanner is not None:
            self.scanner.terminate()


def initialize_interrupt_handling(command: ScanCommand) -> None:

    def handle_interrupt(signal_number: int, stack) -> None:
        revert_progress_changes()
        if parent_process() is None:
            log.info('Scan command interrupted, stopping...')
            command.terminate()
            reset_terminal()
        sys.exit(130)

    signal.signal(signal.SIGINT, handle_interrupt)


def print_error(message: str) -> None:
    if sys.stderr is not None:
        print(message, file=sys.stderr)
    else:
        print(message)


def reset_terminal_with_error(message: str) -> None:
    reset_terminal()
    print_error(message)


def display_version() -> None:
    print(f"Wordfence CLI {__version__}")
    jit_support_text = 'Yes' if pcre.HAS_JIT_SUPPORT else 'No'
    print(f"PCRE Version: {pcre.VERSION} - JIT Supported: {jit_support_text}")


def main(config) -> int:
    command = None
    try:
        if config.version:
            display_version()
            return 0
        if config.quiet:
            log.setLevel(logging.CRITICAL)
        elif config.debug:
            log.setLevel(logging.DEBUG)
        elif config.verbose or (
                    config.verbose is None
                    and sys.stdout is not None and sys.stdout.isatty()
                ):
            log.setLevel(logging.INFO)
        configurer = Configurer(config)
        configurer.check_config()
        if not config.configure:
            command = ScanCommand(config)
            initialize_interrupt_handling(command)
            command.execute()
        return 0
    except api.licensing.LicenseRequiredException:
        reset_terminal_with_error('A valid Wordfence CLI license is required')
        return 1
    except BaseException as exception:
        if command is not None:
            command.terminate()
        reset_terminal()
        if isinstance(exception, ExceptionContainer):
            if config.debug:
                print_error(exception.trace)
                return 1
            exception = exception.exception
        if config.debug:
            raise exception
        else:
            if isinstance(exception, ProgressException):
                print_error(
                        'The current terminal size is inadequate for '
                        'displaying progress output for the current scan '
                        'options'
                    )
            elif isinstance(exception, SystemExit):
                raise exception
            else:
                print_error(f'Error: {exception}')
        return 1



import fcntl
import os
from typing import Optional, IO, TextIO
from enum import IntEnum


class IoException(Exception):
    pass


class StreamReader:

    def __init__(self, stream: TextIO, delimiter: str, chunk_size: int = 1024):
        self.stream = stream
        self.delimiter = delimiter
        self.chunk_size = chunk_size
        self._buffer = ''
        self._end_of_stream = False

    def read_entry(self) -> Optional[str]:
        while True:
            index = self._buffer.find(self.delimiter)
            if index != -1:
                entry = self._buffer[:index]
                self._buffer = self._buffer[index + 1:]
                return entry
            elif not self._end_of_stream:
                read = self.stream.read(self.chunk_size)
                if read == '':
                    self._end_of_stream = True
                self._buffer += read
            else:
                break
        if len(self._buffer) > 0:
            path = self._buffer
            self._buffer = ''
            return path
        else:
            return None


class LockType(IntEnum):
    EXCLUSIVE = fcntl.LOCK_EX
    SHARED = fcntl.LOCK_SH


class FileLock:

    def __init__(self, file: IO, lock_type: LockType = LockType.EXCLUSIVE):
        self.file = file
        self.lock_type = lock_type

    def _operate_lock(self, action: int):
        fcntl.flock(self.file, action)

    def __enter__(self):
        self._operate_lock(self.lock_type.value)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._operate_lock(fcntl.LOCK_UN)


def resolve_path(path: str) -> str:
    """ Resolve a path to a normalized, absolute path """
    return os.path.abspath(os.path.expanduser(path))


def ensure_directory_is_writable(path: str, create_mode: int = 0o700) -> str:
    """ Ensure that the specified directory is writable, """
    """ creating it and parent directories as needed. Note """
    """ that the checks here are not atomic; this assumes """
    """ that nothing else is modifying the filesystem which """
    """ is not guaranteed. """
    path = resolve_path(path)
    if os.path.exists(path):
        if os.path.isdir(path):
            if not os.access(path, os.W_OK):
                raise IoException(f'Directory at {path} is not writable')
            if not os.access(path, os.X_OK):
                raise IoException(f'Directory at {path} is not executable')
            return path
        else:
            raise IoException(f'Path {path} exists and is not a directory')
    try:
        os.makedirs(path, mode=create_mode, exist_ok=True)
    except OSError:
        raise IoException('Failed to create directory at {$path}')
    return path
